/*
 * Po drate 4 Lemmings clone
 * Copyright (C) 2009 Jiri Benc
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include "lemmings.h"

#define MAX_STEPS	300
#define MAX_FALL	3

char level[YSIZE][XSIZE + 1];
static int start_cnt = 20;
static int start_x = 13, start_y = 10;
static int dest_cnt = 18;
static int delay = 3;

static char updated[YSIZE][XSIZE + 1];
static struct lemming *lemmings;
static int lem_alive, lem_saved;
static int steps;

void fatal(char *msg)
{
	fprintf(stderr, "Error: %s\n", msg);
	exit(1);
}

static void read_level(char *name)
{
	FILE *f;
	int i, j, len;
	char buf[256], *s;

	f = fopen(name, "r");
	if (!f)
		fatal("open level");
	fgets(buf, 256, f);
	for (s = buf; *s && (*s != '/'); s++)
		;
	if (*s)
		*s++ = '\0';
	else
		s = buf;
	start_cnt = atoi(s);
	dest_cnt = atoi(buf);
	for (i = 0; i < YSIZE; i++) {
		fgets(buf, 256, f);
		strncpy(level[i], buf, XSIZE);
		len = strlen(level[i]);
		for (j = 0; j < len; j++)
			if (level[i][j] == '@') {
				start_x = j;
				start_y = i;
				level[i][j] = ' ';
			}
	}
	fclose(f);
}

static void screen_update_start(void)
{
	memset(updated, 0, sizeof(updated));
}

static void __screen_update(int x, int y, char c)
{
	printf("\x1b[%d;%dH%c", y + 1, x + 1, c);
}

static void screen_update(int x, int y, char c, int transparent)
{
	if (!updated[y][x]) {
		__screen_update(x, y, c);
		if (!transparent)
			updated[y][x] = 1;
	}
}

static void screen_update_end(void)
{
	fputs("\x1b[25;1H", stdout);
	fflush(stdout);
}

static void screen_rewrite(void)
{
	int i, j;
	char c;

	fputs("\x1b[H\x1b[2J", stdout);
	for (i = 0; i < YSIZE; i++) {
		for (j = 0; j < XSIZE; j++) {
			c = level[i][j];
			if (c == '%')
				c = ' ';
			putchar(c);
		}
		putchar('\n');
	}
}

void draw_lemming(struct lemming *lem)
{
	char *draw;

	draw = lem->draw;
	if (!draw || !lem->skill)
		draw = "pq";
	screen_update(lem->x, lem->y, lem->dir ? draw[1] : draw[0], 0);
}

void erase_lemming(struct lemming *lem)
{
	screen_update(lem->x, lem->y, level[lem->y][lem->x], 1);
}

void move_lemming(struct lemming *lem, int xrel, int yrel)
{
	erase_lemming(lem);
	lem->x += xrel;
	lem->y += yrel;
	draw_lemming(lem);
}

void highlight_lemming(struct lemming *lem, char c)
{
	__screen_update(lem->x, lem->y, c);
}

void modify_level(int x, int y, char c)
{
	level[y][x] = c;
	__screen_update(x, y, c);
}

static void activate_skill(struct lemming *lem, int sk)
{
	if (lem->skill && sk < SKILLS)
		return;		// it's not allowed to change the skill, hehehe
	lem->skill = skills + sk;
	lem->skill_data = 0;
	lem->draw = NULL;
	if (lem->skill->activate && !lem->skill->activate(lem)) {
		highlight_lemming(lem, '!');
		lem->skill = NULL;
	}
}

static void add_lemming(int x, int y, int dir)
{
	struct lemming *lem;

	lem = calloc(1, sizeof(struct lemming));
	if (!lem)
		fatal("out of memory");
	lem->x = x;
	lem->y = y;
	lem->dir = dir;
	if (!lemmings) {
		lem->prev = lem->next = lem;
		lemmings = lem;
	} else {
		lem->next = lemmings;
		lem->prev = lemmings->prev;
		lemmings->prev = lem;
		lem->prev->next = lem;
	}

	lem_alive++;
	draw_lemming(lem);
}

void delete_lemming(struct lemming *lem)
{
	if (lem->next == lem)
		lemmings = NULL;
	else {
		lem->next->prev = lem->prev;
		lem->prev->next = lem->next;
		if (lemmings == lem)
			lemmings = lem->next;
	}

	lem_alive--;
	erase_lemming(lem);
	free(lem);
}

void kill_lemming(struct lemming *lem)
{
	activate_skill(lem, SKILLS);
}

static int check_blocker(struct lemming *lem, int xrel, int yrel)
{
	if (!BLOCKER(lem->x + xrel, lem->y + yrel))
		return 0;
	lem->dir = !lem->dir;
	draw_lemming(lem);
	return 1;
}

static void do_step_lemming(struct lemming *lem)
{
	int xrel;

	if (FINISH(lem->x, lem->y)) {
		delete_lemming(lem);
		lem_saved++;
		return;
	}
	if (lem->skill && lem->skill->move) {
		if (lem->skill->move(lem))
			return;
	}
	if (lem->y == YSIZE - 2) {
		// kill lemming unconditionally when it reaches the bottom
		// of the screen
		erase_lemming(lem);
		lem->y++;
		kill_lemming(lem);
		return;
	}
	if (lem->y < YSIZE - 2 && EMPTY(lem->x, lem->y + 1)) {
		// falling
		if (lem->skill && lem->skill->pre_fall)
			if (!lem->skill->pre_fall(lem))
				return;
		lem->fall++;
		move_lemming(lem, 0, 1);
		if (lem->skill && lem->skill->post_fall)
			lem->skill->post_fall(lem);
		return;
	}
	if (lem->fall > MAX_FALL) {
		// fell from a too big height
		erase_lemming(lem);
		lem->y++;
		kill_lemming(lem);
		return;
	}
	lem->fall = 0;
	if (!lem->dir && lem->x == XSIZE - 2) {
		// kill lemming at the right side of the screen
		erase_lemming(lem);
		lem->x++;
		kill_lemming(lem);
		return;
	}
	if (lem->dir && lem->x == 1) {
		// and at the left side
		erase_lemming(lem);
		lem->x--;
		kill_lemming(lem);
		return;
	}

	xrel = DIR_TO_XREL(lem);
	if (!EMPTY(lem->x + xrel, lem->y)) {
		if (!EMPTY(lem->x + xrel, lem->y - 1)) {
			if (lem->skill && lem->skill->wall) {
				lem->skill->wall(lem);
				return;
			}
			lem->dir = !lem->dir;
			draw_lemming(lem);
			return;
		}
		if (check_blocker(lem, xrel, -1))
			return;
		move_lemming(lem, xrel, -1);
		return;
	}
	if (EMPTY(lem->x + xrel, lem->y + 1)) {
		if (check_blocker(lem, xrel, 1))
			return;
		move_lemming(lem, xrel, 1);
		return;
	}
	if (check_blocker(lem, xrel, 0))
		return;
	move_lemming(lem, xrel, 0);
}

static void do_step(void)
{
	struct lemming *lem, *last, *next;
	int skill;

	screen_update_start();
	if (lemmings) {
		last = lemmings->prev;
		lem = lemmings;
		while (1) {
			next = lem->next;
			do_step_lemming(lem);
			if (lem == last)
				break;
			lem = next;
		}
	}
	if (!(steps % 2) && start_cnt) {
		add_lemming(start_x, start_y, 0);
		start_cnt--;
	}

	advance_program();
	if (lemmings) {
		last = lemmings->prev;
		lem = lemmings;
		while (1) {
			skill = program_get_skill();
			if (skill < 0)
				break;
			next = lem->next;
			if (skill > 0)
				activate_skill(lem, skill - 1);
			if (lem == last)
				break;
			lem = next;
		}
	}

	steps++;
	screen_update_end();
	usleep(delay * 100 * 1000);
}

void adjust_speed(int rel)
{
	delay += rel;
	if (delay < 1)
		delay = 1;
	else if (delay > 5)
		delay = 5;
}

int odd_step(void)
{
	return steps & 1;
}

static void print_password(void)
{
	puts("Heslo je krecek.");
}

static void terminate(int param)
{
	(void)param;
	exit(1);
}

int main(void)
{
	signal(SIGPIPE, terminate);
	read_level("level");
	get_program();
	screen_rewrite();
	while (steps < MAX_STEPS) {
		do_step();
		if (!lem_alive)
			break;
	}
	if (lem_saved >= dest_cnt)
		print_password();
	return 0;
}
