/*
 * Po drate 4 Lemmings clone
 * Copyright (C) 2009 Jiri Benc
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdlib.h>
#include "lemmings.h"

//#define SLOW_BUILDER
//#define SLOW_BASHER
//#define SLOW_MINER
//#define SLOW_DIGGER

static int climber_move(struct lemming *lem)
{
	int dir;

	if (!lem->skill_data)
		return 0;
	dir = -DIR_TO_XREL(lem);	// dir is reversed now, see climber_wall
	if (lem->y > 1 && EMPTY(lem->x, lem->y - 1)) {
		if (!EMPTY(lem->x + dir, lem->y)) {
			move_lemming(lem, 0, -1);
			return 1;
		}
		lem->dir = !lem->dir;	// we're not falling, so restore the original direction
		move_lemming(lem, dir, 0);
		lem->skill_data = 0;
		return 1;
	}
	// let him fall again
	lem->skill_data = 0;
	return 0;
}

static void climber_wall(struct lemming *lem)
{
	// always reverse direction - that way we get a nicer picture of the
	// climbing lemming and don't have to do anything when the lemming
	// falls down again
	lem->dir = !lem->dir;
	lem->skill_data = 1;
	if (!climber_move(lem))
		// just redraw the lemming to turn him around
		draw_lemming(lem);
}

static int floater_pre_fall(struct lemming *lem)
{
	if (!odd_step())
		return 0;
	if (lem->fall > 0)
		lem->draw = "YY";
	return 1;
}

static void floater_post_fall(struct lemming *lem)
{
	lem->draw = NULL;
	if (lem->fall > 1)
		lem->fall = 1;
}

static char exploder_draw[6][3] = { "**", "11", "22", "33", "44", "55" };

static int exploder_activate(struct lemming *lem)
{
	lem->skill_data = 6;
	return 1;
}

static int exploder_move(struct lemming *lem)
{
	if (!lem->skill_data) {
		delete_lemming(lem);
		return 1;
	}
	lem->draw = exploder_draw[--lem->skill_data];
	return 0;
}

static int dying_activate(struct lemming *lem)
{
	lem->draw = "**";
	draw_lemming(lem);
	return 1;
}

static int dying_move(struct lemming *lem)
{
	delete_lemming(lem);
	return 1;
}

static int blocker_activate(struct lemming *lem)
{
	if (EMPTY(lem->x, lem->y + 1))
		return 0;
	modify_level(lem->x, lem->y, 'T');
	delete_lemming(lem);
	return 1;
}

static int builder_activate(struct lemming *lem)
{
	int xrel;

	if (EMPTY(lem->x, lem->y + 1))
		return 0;
	xrel = DIR_TO_XREL(lem);
	if (!CHECK_BORDERS(lem, xrel, 0))
		return 0;
	modify_level(lem->x + xrel, lem->y, '=');
	lem->skill_data = 1;
	return 1;
}

static int builder_move(struct lemming *lem)
{
	int xrel, yrel;

#ifdef SLOW_BUILDER
	if (!odd_step())
		return 1;
#endif

	if (lem->skill_data == 9) {
		lem->skill = NULL;
		return 0;
	}
	xrel = DIR_TO_XREL(lem);
	yrel = -(!((lem->skill_data - 1) % 3));
	if (!CHECK_BORDERS(lem, xrel, yrel) ||
	    !EMPTY(lem->x + xrel, lem->y + yrel)) {
		lem->skill = NULL;
		return 0;
	}
	move_lemming(lem, xrel, yrel);

	yrel = -(!(lem->skill_data % 3));
	modify_level(lem->x + xrel, lem->y + yrel + 1, '=');
	lem->skill_data++;
	return 1;
}

static int basher_activate(struct lemming *lem)
{
	int xrel;

	if (EMPTY(lem->x, lem->y + 1))
		return 0;
	xrel = DIR_TO_XREL(lem);
	if (!CHECK_BORDERS(lem, xrel, 0) ||
	    EMPTY(lem->x + xrel, lem->y))
		return 0;
	lem->draw = "><";
	return 1;
}

static int basher_move(struct lemming *lem)
{
	int xrel;

#ifdef SLOW_BASHER
	if (!odd_step())
		return 1;
#endif

	xrel = DIR_TO_XREL(lem);
	if (!CHECK_BORDERS(lem, xrel, 0)) {
		lem->skill = NULL;
		return 0;
	}
	if (EMPTY(lem->x + xrel, lem->y)) {
		lem->skill = NULL;
		return 0;
	}
	modify_level(lem->x + xrel, lem->y, ' ');
	move_lemming(lem, xrel, 0);

	if (lem->y + 1 < YSIZE - 1) {
		if (!EMPTY(lem->x, lem->y + 1))
			modify_level(lem->x, lem->y + 1, '-');
		else
			lem->skill = NULL;
	}
	if (lem->y - 1 > 0) {
		if (!EMPTY(lem->x, lem->y - 1))
			modify_level(lem->x, lem->y - 1, '-');
	}
	return 1;
}

static void add_walls_around(struct lemming *lem)
{
	if (lem->x - 1 > 0 && !EMPTY(lem->x - 1, lem->y))
		modify_level(lem->x - 1, lem->y, '|');
	if (lem->x + 1 < XSIZE - 1 && !EMPTY(lem->x + 1, lem->y))
		modify_level(lem->x + 1, lem->y, '|');
	if (lem->y + 1 < YSIZE - 1 && !EMPTY(lem->x, lem->y + 1))
		modify_level(lem->x, lem->y + 1, '-');
}

static int miner_activate(struct lemming *lem)
{
	int xrel;

	if (EMPTY(lem->x, lem->y + 1))
		return 0;
	xrel = DIR_TO_XREL(lem);
	if (!CHECK_BORDERS(lem, 0, 1))
		return 0;
	lem->draw = "vv";
	return 1;
}

static int miner_move(struct lemming *lem)
{
	int xrel, yrel;

#ifdef SLOW_MINER
	if (!odd_step())
		return 1;
#endif

	xrel = DIR_TO_XREL(lem) * lem->skill_data;
	yrel = !lem->skill_data;
	lem->skill_data = !lem->skill_data;

	if (!CHECK_BORDERS(lem, xrel, yrel) ||
	    EMPTY(lem->x + xrel, lem->y + yrel) ||
	    EMPTY(lem->x, lem->y + 1)) {
		lem->skill = NULL;
		return 0;
	}
	modify_level(lem->x + xrel, lem->y + yrel, ' ');
	move_lemming(lem, xrel, yrel);
	add_walls_around(lem);
	return 1;
}

static int digger_activate(struct lemming *lem)
{
	if (!CHECK_BORDERS(lem, 0, 1) || EMPTY(lem->x, lem->y + 1))
		return 0;
	lem->draw = "vv";
	return 1;
}

static int digger_move(struct lemming *lem)
{
#ifdef SLOW_DIGGER
	if (!odd_step())
		return 1;
#endif

	if (!CHECK_BORDERS(lem, 0, 1) || EMPTY(lem->x, lem->y + 1)) {
		lem->skill = NULL;
		return 0;
	}
	modify_level(lem->x, lem->y + 1, ' ');
	move_lemming(lem, 0, 1);
	add_walls_around(lem);
	return 1;
}

struct skill skills[SKILLS + 1] = {
	{ .move = climber_move, .wall = climber_wall, }, // climber (a)
	{ .pre_fall = floater_pre_fall, .post_fall = floater_post_fall, }, // floater (b)
	{ .activate = exploder_activate, .move = exploder_move, }, // exploder (c)
	{ .activate = blocker_activate, }, // blocker (d)
	{ .activate = builder_activate, .move = builder_move, }, // builder (e)
	{ .activate = basher_activate, .move = basher_move, }, // basher (f)
	{ .activate = miner_activate, .move = miner_move, }, // miner (g)
	{ .activate = digger_activate, .move = digger_move, }, // digger (h)
	{ .activate = dying_activate, .move = dying_move, },
};
